#
# preamble
#

# import required system modules
import os
import sys
import time
import optparse
import filecmp
import shutil
import subprocess

#
# class definitions
#

# version info class
class git_info(object):
  def __init__(self):
    id = None
    date = None
    branch = None
    tag = None
    author = None
    committer = None
    status = None

# git invocation error exception handler
class GitInvocationError(LookupError):
  pass

#
# helper methods
#

# command line option parsing
def parse_args():
  # usage information
  usage = '%prog src_dir build_dir'

  # define option parser
  parser = optparse.OptionParser(usage=usage)

  # parse command line options
  opts, args = parser.parse_args()

  # check for positional arguments
  if (len(args) != 2):
    parser.error('incorrect number of command line options specified')

  return args[0], args[1]

#
# process management methods
#

# return output from running given command
def call_out(command):
  """
  Run the given command (with shell=False) and return a tuple of
  (int returncode, str output). Strip the output of enclosing whitespace.
  """
  # start external command process
  p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

  # get outputs
  out, _ = p.communicate()

  return p.returncode, out.strip()

def check_call_out(command):
  """
  Run the given command (with shell=False) and return the output as a
  string. Strip the output of enclosing whitespace.
  If the return code is non-zero, throw GitInvocationError.
  """
  # start external command process
  p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

  # get outputs
  out, _ = p.communicate()

  # throw exception if process failed
  if p.returncode != 0:
    raise GitInvocationError('failed to run "%s"' % " ".join(command))

  # convert byte objects to strings, if appropriate
  if not isinstance(out, str) and sys.version_info >= (3,):
    out = str(out, encoding='utf8')

  return out.strip()

#
# git version method
#

def in_git_repository():
  """
  Return True if git is available and we are in a git repository; else
  return False.

  NB: Unfortunately there is a magic number without any documentation to back
  it up. It turns out that git status returns non-zero exit codes for all sorts
  of success conditions, but I cannot find any documentation of them. 128 was
  determined empirically. I sure hope that it's portable.
  """
  ret_code, git_path = call_out(('which', 'git'))
  return (ret_code != 1) and not git_path.startswith('no') \
      and (call_out((git_path, 'status'))[0] != 128)

# generate sed command file containing version info
def generate_git_version_info():
  # info object
  info = git_info()

  git_path = check_call_out(('which', 'git'))

  # determine basic info about the commit
  # %H -- full git hash id
  # %ct -- commit time
  # %an, %ae -- author name, email
  # %cn, %ce -- committer name, email
  git_id, git_udate, git_author_name, git_author_email, \
    git_committer_name, git_committer_email = \
    check_call_out((git_path, 'log', '-1',
      '--pretty=format:%H,%ct,%an,%ae,%cn,%ce')).split(",")

  git_date = time.strftime('%Y-%m-%d %H:%M:%S +0000',
    time.gmtime(float(git_udate)))
  git_author = '%s <%s>' % (git_author_name, git_author_email)
  git_committer = '%s <%s>' % (git_committer_name, git_committer_email)

  # determine branch
  branch_match = check_call_out((git_path, 'rev-parse',
    '--symbolic-full-name', 'HEAD'))
  if branch_match == "HEAD":
    git_branch = None
  else:
    git_branch = os.path.basename(branch_match)

  # determine tag
  status, git_tag = call_out((git_path, 'describe', '--exact-match',
    '--tags', git_id))
  if status != 0:
    git_tag = None

  # refresh index
  check_call_out((git_path, 'update-index', '-q', '--refresh'))

  # check working copy for changes
  status_output = subprocess.call((git_path, 'diff-files', '--quiet'))
  if status_output != 0:
    git_status = 'UNCLEAN: Modified working tree'
  else:
    # check index for changes
    status_output = subprocess.call((git_path, 'diff-index', '--cached',
      '--quiet', 'HEAD'))
    if status_output != 0:
      git_status = 'UNCLEAN: Modified index'
    else:
      git_status = 'CLEAN: All modifications committed'

  # determine version strings
  info.id = git_id
  info.date = git_date
  info.branch = git_branch
  info.tag = git_tag
  info.author = git_author
  info.committer = git_committer
  info.status = git_status

  return info

#
# main program entry point
#

if __name__ == "__main__":
  # parse command line options
  src_dir, build_dir = parse_args()

  # sed filename
  sed_basename = 'git_version.sed'
  sed_tmpfile = '%s/%s.tmp' % (build_dir, sed_basename)
  sed_dstfile = '%s/%s' % (build_dir, sed_basename)

  # change to src_dir
  os.chdir(src_dir)

  # generate version information output, if appropriate
  # NB: First assume that git works and we are in a repository since
  # checking it is expensive and rarely necessary.
  try:
    info = generate_git_version_info()
  except GitInvocationError:
    # We expect a failure if either git is not available or we are not
    # in a repository. Any other failure is unexpected and should throw an
    # error.
    if not in_git_repository():
        sys.exit(0)
    else:
        sys.exit("Unexpected failure in discovering the git version")

  # write temporary sed file
  with open(sed_tmpfile, 'w') as f:
    f.write('s/@ID@/%s/\n' % info.id)
    f.write('s/@DATE@/%s/\n' % info.date)
    f.write('s/@BRANCH@/%s/\n' % info.branch)
    f.write('s/@TAG@/%s/\n' % info.tag)
    f.write('s/@AUTHOR@/%s/\n' % info.author)
    f.write('s/@COMMITTER@/%s/\n' % info.committer)
    f.write('s/@STATUS@/%s/\n' % info.status)

  # update if appropriate
  if os.access(sed_dstfile, os.F_OK) and filecmp.cmp(sed_dstfile, sed_tmpfile):
    os.remove(sed_tmpfile)
  else:
    os.rename(sed_tmpfile, sed_dstfile)

sys.exit()

# vim: syntax=python tw=72 ts=2 et
